/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*  
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/
/*
* Modifications:
*
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

  * Modification author: Hejun Wu (vmnet@cs.ust.hk)
  
********************************************************************************/

#ifndef ATEMU_RADIO_H
#define ATEMU_RADIO_H

#include "../EM/device.h"
#include "../EM/debug.h"
#include "spi.h"
#include "adc.h"

/* CC1000 Configuration registers */
#define RADIO_REG_MAIN      0x00 /* MAIN Register */
#define RADIO_REG_MAIN_RESET_N      0x01
#define RADIO_REG_MAIN_BIAS_PD      0x02
#define RADIO_REG_MAIN_CORE_PD      0x04
#define RADIO_REG_MAIN_FS_PD        0x08
#define RADIO_REG_MAIN_TX_PD        0x10
#define RADIO_REG_MAIN_RX_PD        0x20
#define RADIO_REG_MAIN_F_REG        0x40
#define RADIO_REG_MAIN_RXTX         0x80
#define RADIO_REG_FREQ_2A   0x01 /* Frequency Register 2A */
#define RADIO_REG_FREQ_1A   0x02 /* Frequency Register 1A */
#define RADIO_REG_FREQ_0A   0x03 /* Frequency Register 0A */
#define RADIO_REG_FREQ_2B   0x04 /* Frequency Register 2B */
#define RADIO_REG_FREQ_1B   0x05 /* Frequency Register 1B */
#define RADIO_REG_FREQ_0B   0x06 /* Frequency Register 0B */
#define RADIO_REG_FSEP1     0x07 /* Frequency Seperation Register 1 */
#define RADIO_REG_FSEP0     0x08 /* Frequency Seperation Register 0 */
#define RADIO_REG_CURRENT   0x09 /* Current Consumption Control Register */
#define RADIO_REG_FRONT_END 0x0A /* Front End Control Register */
#define RADIO_REG_PA_POW    0x0B /* PA Output Power Control Register */
#define RADIO_REG_PLL       0x0C /* PLL Control Register */
#define RADIO_REG_PLL_EXT_FILTER    0x80
#define RADIO_REG_PLL_REFDIV        0x78
#define RADIO_REG_PLL_ALARM_DISABLE 0x04
#define RADIO_REG_PLL_ALARM_H       0x02
#define RADIO_REG_PLL_ALARM_L       0x01
#define RADIO_REG_LOCK      0x0D /* LOCK Status Register and signal select */
#define RADIO_REG_LOCK_SELECT       0xF0
#define RADIO_REG_LOCK_SEL_NORMAL        0x00
#define RADIO_REG_LOCK_SEL_CONTINUOUS    0x10
#define RADIO_REG_LOCK_SEL_INSTANT       0x20
#define RADIO_REG_LOCK_SEL_ALARM_H       0x30
#define RADIO_REG_LOCK_SEL_ALARM_L       0x40
#define RADIO_REG_LOCK_SEL_CAL_COMPLETE  0x50
#define RADIO_REG_LOCK_SEL_IF_OUT        0x60
#define RADIO_REG_LOCK_SEL_REFDIV        0x70
#define RADIO_REG_LOCK_SEL_TXPDB         0x80
#define RADIO_REG_LOCK_SEL_MANCHESTER    0x90
#define RADIO_REG_LOCK_SEL_RXPDB         0xA0
#define RADIO_REG_LOCK_SEL_AVG_FILTER    0xD0
#define RADIO_REG_LOCK_SEL_NDIVIDER      0xE0
#define RADIO_REG_LOCK_SEL_FCOMP         0xF0
#define RADIO_REG_LOCK_PLL_ACCURACY 0x08
#define RADIO_REG_LOCK_PLL_LENGTH   0x04
#define RADIO_REG_LOCK_INSTANT      0x02
#define RADIO_REG_LOCK_CONTINUOUS   0x01
#define RADIO_REG_CAL       0x0E /* VCO Calibration Control and Status */
#define RADIO_REG_CAL_START         0x80
#define RADIO_REG_CAL_DUAL          0x40
#define RADIO_REG_CAL_WAIT          0x20
#define RADIO_REG_CAL_CURRENT       0x10
#define RADIO_REG_CAL_COMPLETE      0x08
#define RADIO_REG_CAL_ITERATE       0x07
#define RADIO_REG_MODEM2    0x0F /* Modem Control Register 2 */
#define RADIO_REG_MODEM1    0x10 /* Modem Control Register 1 */
#define RADIO_REG_MODEM1_MLIMIT      0xE0
#define RADIO_REG_MODEM1_MLIMIT_NONE 0x00
#define RADIO_REG_MODEM1_MLIMIT_1    0x20
#define RADIO_REG_MODEM1_MLIMIT_2    0x40
#define RADIO_REG_MODEM1_MLIMIT_3    0x60
#define RADIO_REG_MODEM1_MLIMIT_4    0x80
#define RADIO_REG_MODEM1_MLIMIT_5    0xA0
#define RADIO_REG_MODEM1_MLIMIT_6    0xC0
#define RADIO_REG_MODEM1_MLIMIT_7    0xE0
#define RADIO_REG_MODEM1_LOCK_AVG_IN   0x10
#define RADIO_REG_MODEM1_LOCK_AVG_MODE 0x08
#define RADIO_REG_MODEM1_SETTLING    0x06
#define RADIO_REG_MODEM1_SETTLING_0  0x00
#define RADIO_REG_MODEM1_SETTLING_1  0x02
#define RADIO_REG_MODEM1_SETTLING_2  0x04
#define RADIO_REG_MODEM1_SETTLING_3  0x06
#define RADIO_REG_MODEM1_RESET_N     0x01
#define RADIO_REG_MODEM0    0x11 /* Modem Control Register 0 */
#define RADIO_REG_MODEM0_BAUDRATE   0x70
#define RADIO_REG_MODEM0_DATAFORMAT 0x0C
#define RADIO_REG_MODEM0_FORMAT_NRZ  0x00
#define RADIO_REG_MODEM0_FORMAT_MAN  0x04
#define RADIO_REG_MODEM0_FORMAT_UART 0x08
#define RADIO_REG_MODEM0_XOSCFREQ   0x03
#define RADIO_REG_MATCH     0x12 /* Match Capacitor Array Control Register */
#define RADIO_REG_FSCTRL    0x13 /* Frequency Synthesiser Control Register */
#define RADIO_REG_FSHAPE7   0x14 /* Frequency Shaping Register 7 */
#define RADIO_REG_FSHAPE6   0x15 /* Frequency Shaping Register 6 */
#define RADIO_REG_FSHAPE5   0x16 /* Frequency Shaping Register 5 */
#define RADIO_REG_FSHAPE4   0x17 /* Frequency Shaping Register 4 */
#define RADIO_REG_FSHAPE3   0x18 /* Frequency Shaping Register 3 */
#define RADIO_REG_FSHAPE2   0x19 /* Frequency Shaping Register 2 */
#define RADIO_REG_FSHAPE1   0x1A /* Frequency Shaping Register 1 */
#define RADIO_REG_FSDELAY   0x1B /* Frequency Shaping Delay Register */
#define RADIO_REG_PRESCALER 0x1C /* Prescaler and IF-strip test control */
#define RADIO_REG_TEST6     0x40 /* Test register for PLL LOOP */
#define RADIO_REG_TEST5     0x41 /* Test register for PLL LOOP */
#define RADIO_REG_TEST4     0x42 /* Test register for PLL LOOP */
#define RADIO_REG_TEST3     0x43 /* Test register for VCO */
#define RADIO_REG_TEST2     0x44 /* Test register for Calibration */
#define RADIO_REG_TEST1     0x45 /* Test register for Calibration */
#define RADIO_REG_TEST0     0x46 /* Test register for Calibration */

/* Bit positions in Port D */
#define RADIO_PORTD_PALE    0x01 /* Programming Address Latch Enable */
#define RADIO_PORTD_PCLK    0x04  /* Programming Clock */
#define RADIO_PORTD_PDATA   0x08 /* Programming Data */

/* Synchronization states */
#define CC1000_SYNC_NOSIGNAL   0 /* No signal */
#define CC1000_SYNC_FREQLOW    1 /* Locking onto frequencies, waiting for
* low -> high transition */
#define CC1000_SYNC_FREQHIGH   2 /* Locking onto frequencies, waiting for
* high -> low transition */
#define CC1000_SYNC_FREQLOCK   3 /* Final lock state, waiting for second
* low-> high transition */
#define CC1000_SYNC_CHIPWAIT   4 /* Frequencies locked, waiting for next chip */
#define CC1000_SYNC_CHIPTIME   5 /* Measuring average delay between chips */
#define CC1000_SYNC_LOCKED     6 /* Synchronization complete */
#define CC1000_SYNC_LOST       7 /* Connection lost, reinitialize tuner */




#define WHJ_MAN_TRANSMIT
/* Declare some typedefs */
typedef struct radio_cc1000 radio_cc1000;
typedef error_t (*cc1000_data_callback)(radio_cc1000 *radio, void *pdev,
										pin_drive_enum clock, pin_drive_enum data);
typedef error_t (*cc1000_chpout_callback)(radio_cc1000 *radio, void *pdev,
										  pin_drive_enum chpout);

typedef struct {
    device_obj    device;
	
    /* Pointer to this device's SPI structure */
    struct spi_device* spi;
    spi_drive_func spi_drive_slave;
    spi_drive_func spi_drive_miso;
    spi_drive_func spi_drive_mosi;
    spi_drive_func spi_drive_clk;
    /* Pointer to the ADC callback set-up function */
    adc_set_source_func adc_set_source;
    /* Pointer into the air waves */
    struct radio_air* airwave;
    /* Pointer to the CC1000 object */
    struct radio_cc1000* cc1000;
	
    /* Port A interface */
    register_obj  reg_ddra;
    register_obj  reg_porta;
    register_obj  reg_pina;
    /* Port D interface */
    register_obj  reg_ddrd;
    register_obj  reg_portd;
    register_obj  reg_pind;
    /* Interrupt Control Register (For DOT's INT3) */
    register_obj  reg_eicra;
    /* Interrupt Mask Register (For DOT's INT3) */
    register_obj  reg_eimsk;
    /* Interrupt Flag Register (For DOT's INT3) */
    register_obj  reg_eifr;
	
    /* INT3 (Bound to transmitter's CHP_OUT on DOT) */
    interrupt_obj int_chp;
	
    /* Non-zero if operating in dot-compatibility mode */
    int   dotmode;
	
    /* Port A and D buffers */
    uint8_t porta_ddr;
    uint8_t porta_out;
    uint8_t portd_ddr;
    uint8_t portd_out;
	
    /* Standard deviation of artifical noise to add to the RSSI pin */
    double rssinoise;
} radio_device;

struct radio_cc1000 {
    /* Pointer to the parent's device object */
    device_obj *device;
    /* The CC1000 needs it's own clock */
    clock_obj clk_tick;
    /* Pointer into the air waves */
    struct radio_air* airwave;
	
    /* Optional callbacks */
    cc1000_data_callback func_data;
    void *func_data_dev;
    cc1000_chpout_callback func_chpout;
    void *func_chpout_dev;
	
    /* State of programming pins */
    pin_drive_enum prgm_data;
    /* Previous state of clock pulse */
    int cfg_oldclock, cfg_oldale;
    /* Address being written */
    uint8_t cfg_address;
    /* Data being written or read */
    uint8_t cfg_data;
    /* Position of address/data read/write */
    uint8_t cfg_position;
    /* One if programming, zero if reading */
    uint8_t cfg_writing;
    /* One if reading address, Zero if data */
    uint8_t cfg_readaddress;
    /* One if in an actual configuration cycle */
    uint8_t cfg_active;
	
    /* Radio configuration */
    uint8_t config[128];
    uint32_t freqA;
    uint32_t freqB;
    uint16_t freqSep;
    double freqReceive;
    double freqOne;
    double freqZero;
    /* Reset Status */
    uint8_t stat_inreset;
    /* CPU clock frequency */
    double freq_cpu;
    /* CC1000 Crystal clock frequency */
    double freq_crystal;
    /* Reference Frequency */
    double freq_reference;
	
    /* Current RSSI reading (still in mW) */
    double strength;
    /* Current DIO and DCLK values */
    pin_drive_enum pin_dio, pin_dclk, pin_chpout;
	
    /* Device procrastination */
	/* Calibration Delay */
    uint32_t cal_delay;
	/* Delay after enabling core clock */
    uint32_t core_delay;
	/* Delay after enabling frequency synthesiser */
    uint32_t fs_delay;
	/* Delay after enabling bias generator */
    uint32_t bias_delay;
	
    /* Current transmission status */
	/* TRUE if in an active asynchronous mode */
    uint8_t air_asyncsend;
	/* Current value of bit on the DIO pin */
    uint8_t air_onwire;
	/* Bit as sampled on falling edge */
    int air_transmit;
	/* Bit received on the wire */
    pin_drive_enum air_received;
	/* Current clock value */
    uint32_t air_clock;
	/* Current clock transition period */
    uint32_t air_clockdelta;
	/* Current clock line value */
    uint8_t air_clockout;
	/* Current manchester position */
    uint8_t air_manchester;
	/* Bit rate as according to modem configuration */
    uint32_t air_clockoriginal;
	
    /* Synchronization flags */
	/* TRUE if we are synchronized and can receive data */
    uint8_t sync;
	/* Synchronization state */
    uint8_t sync_state;
	/* Chips left in the synchronization */
    uint8_t sync_chipsleft;
	/* Total chips required for synchronization */
    uint8_t sync_chips;
	/* Number of bits received */
    uint8_t sync_bitcount;
	/* Running cycle total during synchronization */
    uint32_t sync_cyclecount;
	/* Last bit received */
    int sync_bit;
	/* Last frequency seen */
    double sync_lastfreq;
};

typedef struct radio_aircache {
    /* Remote radio location */
    struct radio_air *remote;
	
    /* Are we receiving from them? */
    int transmitting;
    /* Path loss times f^2 */
    double pathloss;
    /* Strength of reception (pre-divided) in dBm */
    double strength;
    /* Frequency it is transmitting */
    double frequency;
    /* Last updated */
    int32_t lastchange;
	
    /* This is a linked list */
    struct radio_aircache* next;
} radio_aircache;

typedef struct radio_air {
    /* Radio position */
    double x;
    double y;
    double z;
	
    /* Noise Level */
    double noiselevel;
    /* Attenuation constant */
    double attenuate;
	
    /* frequency this device is transmitting/receiving on */
    double frequency;
	
    /* non-zero if this device is currently transmitting */
    int transmitting;
    /* Strength of transmission */
    double strength;
	
    /* Threshold of receiption */
    double threshold;
    /* How much frequency needs to change for reception to drop 10dB */
    double deltafreq;
	
    /* Cycle this last changed on (this is a false cycle) */
    int32_t lastchange;
    /* Cycle we last updated our data */
    int32_t lastread;
	
    /* Frequency we expect a one on */
    double freqtrue;
    /* Frequency we expect a zero on */
    double freqfalse;
    /* Schmitt distance, threshold for transitioning frequencies */
    double schmittdist;
    /* Current reception one strength */
    double strengthtrue;
    /* Current reception false strength */
    double strengthfalse;
    /* last bit received */
    double lastfreq;
	
    /* This is a linked list */
    struct radio_air* prev;
    struct radio_air* next;
    /* Links to our cache */
    struct radio_aircache* cache;
} radio_air;

/*****************************************************************************
* The CC1000 emulation bank-end
*/

/* Creates a new CC1000 structure (needs an existing air object and
* requires a clock from the emulator) */
radio_cc1000* cc1000_create_device(radio_air* airwave, device_obj *device);
/* Destroys a device created with cc1000_create_device() */
error_t cc1000_destroy_device(radio_cc1000 *radio);

/* Drive programming pins on chip */
error_t cc1000_program_write(radio_cc1000 *radio, pin_drive_enum clock,
							 pin_drive_enum data, pin_drive_enum pale);
/* Read programming pins on chip (inputs will float) */
error_t cc1000_program_read(radio_cc1000 *radio, pin_drive_enum *clockout,
							pin_drive_enum *dataout, pin_drive_enum *paleout);

/* Drive the data pin on the chip */
error_t cc1000_data_write(radio_cc1000 *radio, pin_drive_enum data);
/* Read the current data and clock on the chip */
error_t cc1000_data_read(radio_cc1000 *radio, pin_drive_enum *dataout,
						 pin_drive_enum *clockout);
/* Set callback on clock changes */
error_t cc1000_data_setcallback(radio_cc1000*, cc1000_data_callback, void*);

/* Read the current state of CHP_OUT */
pin_drive_enum cc1000_chpout_read(radio_cc1000 *radio);
/* Set a callback for notifications on change of CHP_OUT */
error_t cc1000_chpout_setcallback(radio_cc1000*, cc1000_chpout_callback, void*);

/* Read the current value on the RSSI pin as a voltage above GND */
double cc1000_rssi_read(radio_cc1000 *radio);

/*****************************************************************************
* The air emulation back-end for the radio device.
*/

/* Returns a number between 0.0 and 1.0.  Use in a deterministic manner */
double air_random();
/* Returns a gaussian random number with a standard deviation of 1.0 */
double air_gaussian_random();

/* Creates a new node in the airwaves */
radio_air* air_create_device();
/* Destroys a device created with air_create_device() */
error_t air_destroy_device(radio_air *node);
/* Moves the node to a new position in space */
error_t air_set_position(radio_air *node, double x, double y, double z);

/* Sets the power output into the air by this device when transmitting */
error_t air_transmit_power(radio_air *node, double power);
/* Transmits a bit onto the air of frequency frequency */
#ifdef WHJ_MAN_TRANSMIT
error_t air_transmit(radio_air *node, double frequency,char mansend);
#else
error_t air_transmit(radio_air *node, double frequency/*,char mansend*/);
#endif
/* Stops all current transmission */
error_t air_transmit_release(radio_air *node);

/* Returns true if air activity has occured, primarily meant to be an
* optimization */
int air_receive_changed(radio_air *node);

/* Changes the tuner parameters.  The normal 'sense' is centered around
* the frequency and the sensitivity drops 10dB for every 'deltafreq'
* away from this center point.  threshold is used by sense and read in
* order to simplify reception thresholds */
error_t air_receive_tuner(radio_air *node, double frequency, double deltafreq,
						  double threshold);

						  /* Senses the air for the strongest frequency with normal tuning.  Returns
						  * true if a frequency has been sensed, also filling in the frequency and
						  * strength parameters if provided.  If false is returned, no frequency of
* necessary strength was found and frequency and strength are undefined. */
int air_receive_sense(radio_air *node, double *frequency, double *strength);

/* Sets the frequencies and schmitt threshold for true and false.  On a
* read, the value changes if the frequency passes within schmitt of the
* opposite value.  The schmitt distance is a ratio, and must be > .5
* This acts to help eliminate glitches while transitioning in a noisy
* environment. */
error_t air_receive_frequencies(radio_air *node, double freqfalse,
								double freqtrue, double schmittdist);

								/* Performs the actual bit reception.  Each bit frequency is tuned
								* seperately and a bit is received if either value exceeds threshold.
								* If a bit of sufficent strength is received, true is returned, otherwise
								* false.  In all cases, bit and strength are filled to give a best
* representation of what is on the wire. */
int air_receive(radio_air *node, int *bitout, double *strengthout);

#endif /* ATEMU_TIMER8_H */
