/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#ifndef ATEMU_TIMER16_H
#define ATEMU_TIMER16_H

#include "../EM/device.h"

#define TIMER16_CONTROLA_COM1A1 0x80
#define TIMER16_CONTROLA_COM1A0 0x40
#define TIMER16_CONTROLA_COM1A (TIMER16_CONTROLA_COM1A1|TIMER16_CONTROLA_COM1A0)
#define TIMER16_CONTROLA_COM1B1 0x20
#define TIMER16_CONTROLA_COM1B0 0x10
#define TIMER16_CONTROLA_COM1B (TIMER16_CONTROLA_COM1B1|TIMER16_CONTROLA_COM1B0)
#define TIMER16_CONTROLA_COM1C1 0x08
#define TIMER16_CONTROLA_COM1C0 0x04
#define TIMER16_CONTROLA_COM1C (TIMER16_CONTROLA_COM1C1|TIMER16_CONTROLA_COM1C0)
#define TIMER16_CONTROLA_WGM11  0x02
#define TIMER16_CONTROLA_WGM10  0x01
#define TIMER16_CONTROLA_WGM1  (TIMER16_CONTROLA_WGM11|TIMER16_CONTROLA_WGM10)

#define TIMER16_CONTROLB_ICNC1  0x80
#define TIMER16_CONTROLB_ICES1  0x40
#define TIMER16_CONTROLB_WGM13  0x10
#define TIMER16_CONTROLB_WGM12  0x08
#define TIMER16_CONTROLB_WGM1  (TIMER16_CONTROLB_WGM13|TIMER16_CONTROLB_WGM12)
#define TIMER16_CONTROLB_CS12   0x04
#define TIMER16_CONTROLB_CS11   0x02
#define TIMER16_CONTROLB_CS10   0x01
#define TIMER16_CONTROLB_CS1 \
(TIMER16_CONTROLB_CS12|TIMER16_CONTROLB_CS11|TIMER16_CONTROLB_CS10)
#define TIMER16_CONTROLB_RESERVED 0x20

#define TIMER16_CONTROLC_FOC1A  0x80
#define TIMER16_CONTROLC_FOC1B  0x40
#define TIMER16_CONTROLC_FOC1C  0x20
#define TIMER16_CONTROLC_RESERVED 0x1F

/* Timer1 is on both mask registers */
#define TIMER16_MASK_TICIE1     0x08
#define TIMER16_MASK_OCIE1A     0x04
#define TIMER16_MASK_OCIE1B     0x02
#define TIMER16_MASK_TOIE1      0x01
#define TIMER16_EMASK_OCIE1C    0x01
/* Timer3 is only on Extended Mask */
#define TIMER16_EMASK_TICIE3    0x10
#define TIMER16_EMASK_OCIE3A    0x08
#define TIMER16_EMASK_OCIE3B    0x04
#define TIMER16_EMASK_TOIE3     0x02
#define TIMER16_EMASK_OCIE3C    0x01

/* Special Function IO Register */
#define TIMER16_SFIOR_TSM       0x80
#define TIMER16_SFIOR_PSR       0x01

/* Clock modes */
enum {
    /* Normal counting */
    TIMER16_MODE_NORMAL,
		/* Clear Timer on Compare (to TOP) */
		TIMER16_MODE_CTC,
		/* Fast PWM */
		TIMER16_MODE_FASTPWM,
		/* Phase Correct PWM */
		TIMER16_MODE_PHASEPWM,
		/* Frequency and Phase Correct PWM */
		TIMER16_MODE_FREQPWM
};

typedef struct {
    device_obj    device;
	
    register_obj  reg_controlA;
    register_obj  reg_controlB;
    register_obj  reg_controlC;
    register_obj  reg_count;
    register_obj  reg_compareA;
    register_obj  reg_compareB;
    register_obj  reg_compareC;
    register_obj  reg_capture;
    register_obj  reg_highbyte;
    register_obj  reg_intmask;
    register_obj  reg_intemask;
    register_obj  reg_intflag;
    register_obj  reg_inteflag;
    register_obj  reg_sync;
	
    clock_obj     clk_tick;
    interrupt_obj int_overflow;
    interrupt_obj int_compareA;
    interrupt_obj int_compareB;
    interrupt_obj int_compareC;
    interrupt_obj int_input;
	
    /* For register reads, the top 8 bits is cached in a shared register */
    uint8_t  temp;
    /* The count */
    uint16_t count;
    uint16_t capture;
    /* The compare registers (buffered) */
    uint16_t compareA;
    uint16_t compareAbuf;
    uint16_t compareB;
    uint16_t compareBbuf;
    uint16_t compareC;
    uint16_t compareCbuf;
    /* Control registers */
    uint8_t controlA;
    uint8_t controlB;
    uint8_t controlC;
    /* Interrupt masks and flags */
    unsigned intCompareA:1;
    unsigned maskCompareA:1;
    unsigned lastCompareA:1;
    unsigned intCompareB:1;
    unsigned maskCompareB:1;
    unsigned lastCompareB:1;
    unsigned intCompareC:1;
    unsigned maskCompareC:1;
    unsigned lastCompareC:1;
    unsigned intOverflow:1;
    unsigned maskOverflow:1;
    unsigned lastOverflow:1;
    unsigned intCapture:1;
    unsigned maskCapture:1;
    unsigned lastCapture:1;
	
    /* Are we using only emask (Timer3) or both emask and mask (Timer1) */
    unsigned emaskonly:1;
    /* Timer is asleep */
    unsigned sleeping:1;
    /* Buffer compare writes */
    unsigned buffercompare:1;
    /* Compares are disabled the cycle after being written */
    unsigned skipnextcompA:1;
    unsigned skipnextcompB:1;
    unsigned skipnextcompC:1;
    /* Timer Sync Mode and Prescaler Reset */
    unsigned enabletsm:1;
    unsigned enablepsr:1;
    /* True when running on external clocking */
    unsigned externalclock:1;
    /* Last tick from the prescaler */
    unsigned lasttick:1;
    /* Counting direction for Phase Correct PWM */
    unsigned direction:1;
    /* Clock divide */
    uint16_t prescaler_divisor;
    uint16_t clock_divide;
    /* Pointer to clock top */
    uint16_t *top;
    /* Pointer to overflow id */
    uint16_t *overflow;
    /* Value to hold "constant" maxes */
    uint16_t max;
    uint16_t bottom;
    /* Current clock mode */
    int mode;
	
    /* External clock driving */
    pin_drive_enum externalvalue;
    pin_drive_enum externalprevious;
} timer16_device;

/* The external clock generation pin is handled by external hardware (or
* possibly software driven).  In any case, it is left up to external hardware
* to drive this pin as appropriate. */
error_t timer16_drive_clock(timer16_device* device, pin_drive_enum value);
/* typedef, for type safety encouragement in dependant modules */
typedef error_t (*timer16_drive_func)(timer16_device*, pin_drive_enum);

#endif /* ATEMU_TIMER16_H */
